-- Model Name: 	Climate Change Model
--			
-- Author:	Srinivasa Raghavendra Bhuvan Gummidi & Milan Secerov

-----------------------------------------------------------------
-- PARAMETERS
-----------------------------------------------------------------

--Initial Temperature of Earth
INITIAL_TEMPERATURE = 12
--Temperature where IR generation starts
IR_GEN_TEMPERATURE = 20
--temperature coefficient used in the formula to calculate temperature
TEMPERATURE_COEF = 0.05 --try varying this value original value is 0.01
--Albedo Value of entire Earth
ALBEDO_EARTH = 0.3
--Albedo Value of Clouds
ALBEDO_CLOUDS = 0.7
--Add clouds by setting the value of cloud flag as 0
CLOUD_FLAG = 0
--cloud speed value to control the movement of clouds
RANDOM_CLOUDS = 0.5
--probability of converting into IRs
PROB_IR_GEN = 0.5
--random value generator
rand = Random{seed=1}
--y value of sky starting
START_SKY = 0
--y value of sky ending and earthsurface starting
END_SKY = 70
--y value of earth surface ending
END_EARTHSURFACE = 76
--y value of earth  ending
END_EARTH = 99
--max X value
MAXVALUEX = 124
--INCREMENT value
INCREMENT_VALUE = 4
--number of co2 molecules
NUM_CO2_MOLECULES = 25 --set this value to 0 if there are no CO2 molecules
--CO2 Dimensions
CO2_X = 122
CO2_Y = 55

-----------------------------------------------------------------
--Cell Variable cover Values
-----------------------------------------------------------------
EARTH_VAL = "earth"
SKY_VAL = "sky"
EARTHSURFACE_VAL = "earthSurface"
CLOUDS_VAL = "clouds"
RAYS_VAL = "rays"
IRS_VAL = "IRs"
CO2_VAL = "co2"
DOTS_VAL = "dots"

-----------------------------------------------------------------
--Global Variables
-----------------------------------------------------------------
--i=76
dotRCount=78
irCounter=74
initXval = 0
initCount = 0
numHeats = 0
numIRs = 0
set_min_cloudX=0
set_max_cloudX=10
set_min_cloudY=25
set_max_cloudY=32
counter=0
countMapCo2 = 0
IRRefractFlag =0

-----------------------------------------------------------------
--MODEL
-----------------------------------------------------------------

-- Cellular space with dimensions and temperature attribute
cs = CellularSpace {
	xdim = 125,
	ydim = 100,
	temperature = INITIAL_TEMPERATURE
	
}

--Creating neighborhood for the cellular space cs
cs:createNeighborhood{
	strategy = "moore",
	self = false
}

--Dividing the Cellular space into thre components, earth, earth surface and sky
forEachCell(cs, function(cell)
	if cell.y > END_SKY and cell.y <=END_EARTHSURFACE then  
        cell.cover = EARTHSURFACE_VAL
    elseif cell.y>END_EARTHSURFACE then
        cell.cover = EARTH_VAL  
    else
        cell.cover = SKY_VAL  
    end
end)

--Synchronizing the cellular space
cs:synchronize()

--Function to create the Co2 molecules in the cellular space if there are any co2 molecules
function createCO2(numco2)
	--checking whether there are any co2 molecules in the system
	if numco2 >0 then
		for co2Count = 0,numco2 do
			--placing the co2 molecules in the cellular space randomly in the sky
			coordco2 = Coord{x =  rand:integer(CO2_X), y = rand:integer(CO2_Y)}
			cellco2 = cs:getCell(coordco2)
			cellco2.cover = CO2_VAL
		end
	end
end

--Creating the co2 molecules at the inital stage
createCO2(NUM_CO2_MOLECULES)

--Function to run the model by initially creating sun rays and subsequent interactions with the different objects in the sky
function runModel(self)
	--creating the clouds and move the co2 molecules around the cellular space randomly
	--the random number is used to slow down the movement, so that its perceivable
	if rand:number(0,1)>RANDOM_CLOUDS then
		--check whether clouds are needed
		if CLOUD_FLAG ==0 then
			--create and move the clouds in the atmosphere
			createClouds()
			--Checks whether there are co2 molecules as part of model
			if NUM_CO2_MOLECULES>0 then
				--moves the co2 molecules around there neighborhood
				moveCo2()
				--used to counter the movement of clouds
				countBalanceCO2()
			end
		end
	end
	--checks whether the ray has reached the earthsurface end or not
	if initCount<=END_EARTHSURFACE then
		--creates sun rays till reaches the earth surface
		coord2 = Coord{x = initXval , y = initCount}
		cell1= cs:getCell(coord2)
		cell1.cover = RAYS_VAL
			--block to visualize the movement of the sun rays and removing the trailing cells after 2 time steps
			if initXval>1 and initCount>4 then
				coord3 = Coord{x = initXval-2 , y = initCount-8}
				cell5= cs:getCell(coord3)
				if cell5.cover ~= CLOUDS_VAL then
					cell5.cover = SKY_VAL
				end
			end
		--if the sun ray encounters a cloud, then the probability of its reflection is 	ALBEDO_CLOUDS
		if cell1.past.cover == CLOUDS_VAL and rand:number(0,1)>ALBEDO_CLOUDS then
			reflectClouds()
		end
		-- increment counters
		initXval = initXval + 1
		initCount = initCount + 4
	end
	-- setting values to parameters of function
	initReflection = initXval
	initRefraction = initXval
	--based on the probability of the ray with the value of  ALBEDO_EARTH, 
	--the incoming sun ray would reflect back to sky or gets absorbed by the earth
	if rand:number(0,1)>ALBEDO_EARTH then
		dotsRefraction(initRefraction,self)
	else
		raysReflection(initReflection)
	end
end

--Function to maintain the number of co2 molecules to counter the movement of clouds
function countBalanceCO2()
	local countCo2 =0
	--counts the number of co2 molecules at a given instant
	forEachCell(cs,function(cell)
		if cell.cover == CO2_VAL then
			countCo2 =countCo2 + 1
		end
	end)
	diffCo2 = NUM_CO2_MOLECULES - countCo2
	--if the difference is greater than 0, then the difference is generated
	if diffCo2>0 then
		createCO2(diffCo2)
	end
end

--Function to move the co2 molecules around in the cellular space
function moveCo2()
	forEachCell(cs,function(cell)
		if cell.cover == CO2_VAL and cell.y<70 then
			newCo2 = cell:getNeighborhood():sample()
			if newCo2.cover ~= CLOUDS_VAL then
				newCo2.cover = CO2_VAL
			end
			cell.cover = SKY_VAL
		end
	end)
end

--Function to reflect the incoming sun rays from space by the clouds
function reflectClouds()
	--tracing the path of the reflected incoming sun ray
	for reflectClCount = cell1.y, 0,-4 do
			coordref = Coord{x = initReflection , y = reflectClCount}
			--checks whether the path crosses the boundaries of cellular space or not
			if initReflection<=MAXVALUEX and initReflection >0 then
				cellClRef= cs:getCell(coordref)
				cellClRef.cover = RAYS_VAL
			end	
			reflectClCount=reflectClCount-4
			initReflection = initReflection+1
	end
	--resetting the values for the next time step
	initXval = rand:integer(30)
	initCount = 0
end

--Function to reflect the reflected sun rays from earth surface by the clouds
function reflectCloudsEarth()
	initReflection = self2.x
	--tracing the path of the reflected sun ray coming from earth surface
	for reflectCECount = self2.y, END_EARTHSURFACE,4 do
			coordref2 = Coord{x = initReflection , y = reflectCECount}
			if initReflection<=MAXVALUEX and initReflection >0 then
				cellClRef2= cs:getCell(coordref2)
				cellClRef2.cover = RAYS_VAL
			end	
			initReflection = initReflection + 1
	end
	--based on the probability of the ray with the value of  ALBEDO_EARTH, 
	--the reflected sun ray would reflect back to sky or gets absorbed by the earth
	if rand:number(0,1)<=ALBEDO_EARTH then
		raysReflection(initReflection)
	else
		dotsRefraction(initReflection,self)
	end
end

--Function to check the temperature of the earth and if it crosses IR_GEN_TEMPERATURE then generate Infrared from earth 
function checkTemperature()
	-- Check whether the current temperature is more than IR_GEN_TEMPERATURE
	if cs.temperature >IR_GEN_TEMPERATURE then
		--the probability of generating the infrared is PROB_IR_GEN
		--the number of escaping infrareds should be always less than the number of heats absorbed by the earth
		if rand:number(0,1)<=PROB_IR_GEN then-- and numIRs<numHeats then
			numIRs = numIRs+1
			--Flag to not negate the temperature growth in case IR escapes
			IRPresent = 1
			initRefraction = initRefraction+rand:integer(30,50)
			--tracing the path of the escaping infrared ray
			while irCounter>0 do
				coord12 = Coord{x = initRefraction , y = irCounter}
				coord12 = Coord{x = initRefraction , y = irCounter}
				self6= cs:getCell(coord12)
				--flag to check whether there is any co2 molecule in its neighborhood
				local flagIR =0
				forEachNeighbor(self6,function(cell,neighbor2)
					if neighbor2.cover == CO2_VAL or self6.cover ==CO2_VAL then
						reflectIRs()
						flagIR =1
					end
				end)
				--if there are any co2 molecules in the neighborhood then reverse the infrared ray
				if flagIR ~=0 then
					break
				end
				self6.cover = IRS_VAL
				irCounter = irCounter-4
				initRefraction = initRefraction+1
			end
			--print("numIRs",numIRs)
		end
		--resetting the value of the counter for next time step
		irCounter=END_EARTHSURFACE
	end
end

--Function to reflect the IRs when it encounters a co2 molecule in its neighborhood
function reflectIRs()
	initReflection = self6.x
	--tracing the path of the reflected IR from the co2 molecule to the earth
	for reflectIRCount = self6.y, END_EARTHSURFACE,4 do
			coordref3 = Coord{x = initReflection , y = reflectIRCount}
			if initReflection<=MAXVALUEX and initReflection >0 then
				cellClRef3= cs:getCell(coordref3)
				cellClRef3.cover = IRS_VAL
			end	
			initReflection = initReflection + 1
	end
	--flag to know, whether it is necessary to check temperature or not
	IRRefractFlag =1
	--all the infrared rays reflected towards earth are absorbed by the earth
	dotsRefraction(initReflection,self)
end

--Function to trace the path of the reflect sun rays by the earth surface
function raysReflection(initReflection)
	--checking whether the ray touched the earth surface or not
	if cell1.y == END_EARTHSURFACE then
		--tracing the path of the reflected ray
		for reflectCount = END_EARTHSURFACE, 0,-4 do
			coord5 = Coord{x = initReflection , y = reflectCount}
			if initReflection<=MAXVALUEX and initReflection >0 then
				self2= cs:getCell(coord5)
				self2.cover = RAYS_VAL
				--if the reflected ray from earth encounters a cloud, then it has a probability to penetrate clouds of ALBEDO_CLOUD
				if self2.past.cover == CLOUDS_VAL and rand:number(0,1)>ALBEDO_CLOUDS then
					--reflecting the rays that are not penetrated into the clouds
					reflectCloudsEarth(self2)
					break
				end
			end	
			reflectCount=reflectCount-4
			initReflection = initReflection+1
		end
	end
end

--Function to trace the path of absorbed sun rays and IRs and count the number of absorbed heats
function dotsRefraction(initRefraction,self)
	if cell1.y == END_EARTHSURFACE then
		--counter for the number of absorbed heats
		numHeats = numHeats + 1
		IRPresent = 0
		--Flag to check whether its IR reflected back to earth or not
		if IRRefractFlag ==0 then
			checkTemperature()
		end
		--calculating the earth's temperature when the heat is absorbed
		if IRPresent ==0 then
			calculateTemperature()
		end
		IRPresent = 1
		--tracing the heat inside the earth
			while dotRCount<END_EARTH do
				coord11 = Coord{x = initRefraction , y = dotRCount}
				if initRefraction<109 and initRefraction >0 then
					self5= cs:getCell(coord11)
					self5.cover = DOTS_VAL
				end
				dotRCount = dotRCount+4
				initRefraction = initRefraction+1
			end
			dotRCount = dotRCount-4
			while dotRCount>END_EARTHSURFACE do
				coord12 = Coord{x = initRefraction , y = dotRCount}
				if initRefraction<109 and initRefraction >0 then
					self11 = cs:getCell(coord12)
					self11.cover = DOTS_VAL
					self10 = self11:getNeighborhood():sample()
					self10.cover = DOTS_VAL
				end
				dotRCount = dotRCount-4
				initRefraction = initRefraction+1
			end
			dotRCount=dotRCount+4
			while dotRCount<END_EARTH do
				coord13 = Coord{x = initRefraction , y = dotRCount}
				if initRefraction<109 and initRefraction >0 then
					self7= cs:getCell(coord13)
					self7.cover = DOTS_VAL
					self11 = self7:getNeighborhood():sample()
					self11.cover = DOTS_VAL
				end
				dotRCount = dotRCount+4
				initRefraction = initRefraction+1
			end
			dotRCount = dotRCount-4
			while dotRCount>END_EARTHSURFACE do
				coord14 = Coord{x = initRefraction , y = dotRCount}
				if initRefraction<109 and initRefraction >0 then
					self8 = cs:getCell(coord14)
					self8.cover = DOTS_VAL
					self9 = self8:getNeighborhood():sample()
					self9.cover = DOTS_VAL
					--self9:move()
				end
				dotRCount = dotRCount-4
				initRefraction = initRefraction+1
			end
			dotRCount=dotRCount+4
			IRRefractFlag =0
	end
end

--Function to calculate the temperature based on the formula given in the Netlogo
function calculateTemperature()
	if cs.temperature >IR_GEN_TEMPERATURE then
		numHeats1 = numHeats - numIRs
	else
		numHeats1 = numHeats
	end
	--This formula is taken from the Netlogo model. . .try varying the value TEMPERATURE_COEF to play with the model
	cs.temperature = 0.99* cs.temperature + (TEMPERATURE_COEF*(INITIAL_TEMPERATURE + 0.1*numHeats1))
	print("Temperature",cs.temperature)
end

--Function to create and move the clouds in the cellular space
function createClouds()
	for cloudX = set_min_cloudX,set_max_cloudX do
		for cloudY = set_min_cloudY,set_max_cloudY do
			coordin1 = Coord{x = cloudX,y = cloudY}
			cloudCell = cs:getCell(coordin1)
			cloudCell.cover = CLOUDS_VAL
		end
	end
	if set_min_cloudX >0 and set_max_cloudX<=MAXVALUEX then
		for cloudX = counter,set_min_cloudX do
			for cloudY = set_min_cloudY,set_max_cloudY do
				coordin1 = Coord{x = cloudX,y = cloudY}
				cloudCell = cs:getCell(coordin1)
				cloudCell.cover = SKY_VAL
			end
		end				
	end	
	set_min_cloudX = set_min_cloudX + 1
	set_max_cloudX = set_max_cloudX + 1
	if set_min_cloudX >0 and set_max_cloudX<=MAXVALUEX-20 then
		for cloudX = set_min_cloudX+20,set_max_cloudX+20 do
			for cloudY = set_min_cloudY+10,set_max_cloudY+10 do
				coordin1 = Coord{x = cloudX,y = cloudY}
				cloudCell = cs:getCell(coordin1)
				cloudCell.cover = CLOUDS_VAL
			end
		end
	end
	if set_min_cloudX >0 and set_max_cloudX<105 then
		for cloudX = counter+20,set_min_cloudX+20 do
			for cloudY = set_min_cloudY+10,set_max_cloudY+10 do
				--print(cloudX,cloudY)
				coordin1 = Coord{x = cloudX,y = cloudY}
				cloudCell = cs:getCell(coordin1)
				cloudCell.cover = SKY_VAL
			end
		end				
	end
	if set_max_cloudX==MAXVALUEX then
		--[[for cloudX = counter,set_min_cloudX do
				for cloudY = set_min_cloudY,set_max_cloudY do
					coordin2 = Coord{x = cloudX,y = cloudY}
					cloudCell1 = cs:getCell(coordin2)
					cloudCell1.cover = SKY_VAL
				end
			end		--]]		
			set_min_cloudX = 5
			set_max_cloudX =15
		end
end

--# Behavior Settings
--Agent used in the model
raysAgent = Agent {
		execute = function(self)
				--Run the model for each time step
				runModel(self)
				--checks whetherit reached the earthsurface and resets the counters
				if initCount>END_EARTHSURFACE then
					initCount = 0
					initXval = rand:integer(40)
				end
			end
			}

--Society with agent "raysAgent" and quantity 
soc = Society{
	instance = raysAgent,
	quantity = 1
}

--Timer
t = Timer{
	Event{action = soc},
	Event{action = cs},	
}

--Environment
e = Environment {
	cs,
	soc,
	t
}

--creating placement with random strategy
e:createPlacement{strategy = "random", max =1}

--Legend
legend1 = Legend {
	grouping = "uniquevalue",
	type = "string",
	colorBar = {
		{value = RAYS_VAL, color = "yellow"},
		{value = DOTS_VAL, color = "red"},
		{value = CLOUDS_VAL, color = "white"},
		{value = IRS_VAL, color = "magenta"},
		{value = EARTH_VAL, color = "green"},
		{value = EARTHSURFACE_VAL, color = "brown"},
		{value = SKY_VAL, color = "blue"},
		{value = CO2_VAL, color = "black"},
	}
}

--Observer for the map
Observer{
	subject = cs,
	attributes = {"cover"},
	legends = {legend1}
}

--Observer for the chart of temp vs time
Observer{
	subject = cs,
	type = "chart",
	attributes = {"temperature"}
}


cs:notify()  -- needed just if you want to see the initial state (before simulation runs)

--Running the model
e:execute(1000)
